@@ -2,7 +2,7 @@
 #   MailScanner - SMTP E-Mail Virus Scanner
 #   Copyright (C) 2002  Julian Field
 #
-#   $Id: MCPMessage.pm,v 1.1.2.9 2004/05/22 12:30:15 jkf Exp $
+#   $Id: MCPMessage.pm,v 1.1.2.21 2005/03/20 13:38:18 jkf Exp $
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -49,7 +49,7 @@ install MIME::Decoder::UU 'uuencode';
 use vars qw($VERSION);
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = substr q$Revision: 1.1.2.9 $, 10;
+$VERSION = substr q$Revision: 1.1.2.21 $, 10;
 
 # Is this message spam? Try to build the spam report and store it in
 # the message.
@@ -76,7 +76,7 @@ sub IsMCP {
   $this->{ismcp} = 0;
   $this->{ishighmcp} = 0;
   $this->{mcpreport} = "";
-  $this->{mcpscore} = 0;
+  $this->{mcpsascore} = 0;
 
   ## If it's a blacklisted address, don't bother doing any checks at all
   #if (MailScanner::Config::Value('spamblacklist', $this)) {
@@ -110,7 +110,7 @@ sub IsMCP {
     $this->{ismcp} = 1;
     $this->{ishighmcp} = 1
       if MailScanner::Config::Value('mcpblacklistedishigh', $this);
-    $this->{spamreport} = $LocalSpamText . ' (' .
+    $this->{mcpreport} = $LocalSpamText . ' (' .
                    MailScanner::Config::LanguageValue($this, 'mcpblacklisted') .
                    ')';
     MailScanner::Log::InfoLog("Message %s from %s (%s) to %s" .
@@ -171,7 +171,7 @@ sub IsMCP {
   my $sascore  = 0;
   ($SAsaysspam, $SAHighScoring, $saheader, $sascore)
     = MailScanner::MCP::Checks($this);
-  $this->{mcpscore} = $sascore; # Save the actual figure for use later...
+  $this->{mcpsascore} = $sascore; # Save the actual figure for use later...
 
   # Fix the return values
   $SAsaysspam = 0 unless $saheader;    # Solve bug with empty SAreports
@@ -232,7 +232,7 @@ sub IsMCP {
                               $this->{from}, $todomain, $ReportText);
   }
 
-  return $this->{isspam};
+  return $this->{ismcp};
 }
     
 
@@ -243,21 +243,44 @@ sub HandleMCP {
   my($this, $HamSpam) = @_;
 
   my($actions, $action, @actions, %actions);
+  my(@extraheaders, $actionscopy, $actionkey);
 
   # Get a space-separated list of all the actions
   if ($HamSpam eq 'nonmcp') {
-    $actions = lc(MailScanner::Config::Value('nonmcpactions', $this));
+    $actions = MailScanner::Config::Value('nonmcpactions', $this);
     # Fast bail-out if it's just the simple "deliver" case that 99% of
     # people will use
     return if $actions eq 'deliver';
   } else {
     # It must be spam as it's not ham
     if ($this->{ishighmcp}) {
-      $actions = lc(MailScanner::Config::Value('highscoremcpactions', $this));
+      $actions = MailScanner::Config::Value('highscoremcpactions', $this);
     } else {
-      $actions = lc(MailScanner::Config::Value('mcpactions', $this));
+      $actions = MailScanner::Config::Value('mcpactions', $this);
     }
   }
+
+  # Find all the bits in quotes, with their spaces
+  $actionscopy = $actions;
+  #print STDERR "Actions = \'$actions\'\n";
+  while ($actions =~ s/\"([^\"]+)\"//) {
+    $actionkey = $1;
+    push @extraheaders, $actionkey;
+    MailScanner::Log::WarnLog("Syntax error in \"header\" action in MCP " .
+                              "actions, missing \":\" in %s", $actionkey)
+      unless $actionkey =~ /:/;
+  }
+  @{$this->{extramcpheaders}} = @extraheaders;
+  $actions = lc($actions);
+  $actions =~ s/^\s*//;
+  $actions =~ s/\s*$//;
+  $actions =~ s/\s+/ /g;
+  #print STDERR "Actions after = \'$actions\'\n";
+  #print STDERR "Extra headers are \"" . join(',',@extraheaders) . "\"\n";
+
+  MailScanner::Log::WarnLog('Syntax error: missing " in MCP actions %s',
+                            $actionscopy) if $actions =~ /\"/;
+
   $actions =~ tr/,//d; # Remove all commas in case they put any in
   @actions = split(" ", $actions);
 
@@ -315,7 +338,8 @@ sub HandleMCP {
       if $HamSpam eq 'spam' && MailScanner::Config::Value('logmcp');
 
     # Mark the message as deleted, so it won't get delivered
-    $this->{deleted} = 1;
+    #$this->{deleted} = 1;
+    $this->{dontdeliver} = 1; # Don't clean or deliver this message, just drop
     # Mark this message as not being delivered by MCP, for later spam filtering
     $this->{mcpdelivering} = 0;
   }
@@ -365,7 +389,11 @@ sub HandleMCP {
     }
     #print STDERR "Storing spam to $spamdir/" . $this->{id} . "\n";
     umask $global::MS->{quar}->{fileumask};
-    $this->{store}->CopyEntireMessage($this, $spamdir, $this->{id});
+    my @paths = $this->{store}->CopyEntireMessage($this, $spamdir, $this->{id});
+    # Remember where we have stored the mcp in an archive, so we never
+    # archive infected messages.
+    push @{$this->{spamarchive}}, @paths;
+    chown $uid, $gid, "$spamdir/" . $this->{id}; # Harmless if this fails
   }
   umask 0077; # Safety net
 
@@ -393,6 +421,7 @@ sub HandleMCPBounce {
 
   my($from,$to,$subject,$date,$spamreport,$hostname);
   my($emailmsg, $line, $messagefh, $filename, $localpostmaster, $id);
+  my($postmastername);
 
   $from = $this->{from};
 
@@ -423,10 +452,11 @@ sub HandleMCPBounce {
   $id = $this->{id};
   $to = join(', ', @{$this->{to}});
   $localpostmaster = MailScanner::Config::Value('localpostmaster', $this);
+  $postmastername  = MailScanner::Config::LanguageValue($this, 'mailscanner');
   $hostname = MailScanner::Config::Value('hostname', $this);
   $subject = $this->{subject};
-  $date = scalar localtime;
-  $spamreport = $this->{spamreport};
+  $date = $this->{datestring}; # scalar localtime;
+  $spamreport = $this->{mcpreport};
 
   # Delete everything in brackets after the SA report, if it exists
   $spamreport =~ s/(spamassassin)[^(]*\([^)]*\)/$1/i;
@@ -467,13 +497,68 @@ sub HandleMCPBounce {
   }
   $messagefh->close();
 
-  # Send the message to the spam sender, but ensure the envelope
-  # sender address is "<>" so that it can't be bounced.
-  $global::MS->{mta}->SendMessageString($this, $emailmsg, '<>')
-    or MailScanner::Log::WarnLog("Could not send sender MCP bounce, %s", $!);
+  if (MailScanner::Config::Value('bouncemcpasattachment', $this)) {
+    $this->HandleMCPBounceAttachment($emailmsg);
+  } else {
+    # Send the message to the spam sender, but ensure the envelope
+    # sender address is "<>" so that it can't be bounced.
+    $global::MS->{mta}->SendMessageString($this, $emailmsg, '<>')
+      or MailScanner::Log::WarnLog("Could not send sender MCP bounce, %s", $!);
+  }
 }
 
 
+# Like encapsulating and sending a message to the recipient, take the
+# passed text as the text and headers of an email message and attach
+# the original message as an rfc/822 attachment.
+sub HandleMCPBounceAttachment {
+  my($this, $plaintext) = @_;
+
+  my $parser = MIME::Parser->new;
+  my $explodeinto = $global::MS->{work}->{dir} . '/' . $this->{id};
+  #print STDERR "Extracting MCP bounce message into $explodeinto\n";
+  my $filer  = MIME::Parser::FileInto->new($explodeinto);
+  $parser->filer($filer);
+
+  my $bounce = eval { $parser->parse_data(\$plaintext) };
+  if (!$bounce) {
+    MailScanner::Log::WarnLog("Cannot parse MCP bounce report, %s", $!);
+    return;
+  }
+  #print STDERR "Successfully parsed bounce report\n";
+
+  # Now make it multipart and push the report into a child
+  $bounce->make_multipart('report');
+
+  # Now turn the original message into a string and attach it
+  my(@original);
+  #my $original = $this->{entity}->stringify;
+  @original = $global::MS->{mta}->OriginalMsgHeaders($this, "\n");
+  push(@original, "\n");
+  $this->{store}->ReadBody(\@original, MailScanner::Config::Value(
+                                          'maxspamassassinsize'));
+
+  $bounce->add_part(MIME::Entity->build(Type        => 'message/rfc822',
+                                        Disposition => 'attachment',
+                                        Top         => 0,
+                                        'X-Mailer'  => undef,
+                                        Data        => \@original));
+
+  # Stringify the message and send it -- this could be VERY large!
+  # Prune all the dead branches off the tree
+  PruneEntityTree($bounce);
+  my $bouncetext = $bounce->stringify;
+  #print STDERR "Spam bounce message is this:\n$bouncetext";
+  if ($bouncetext) {
+    $global::MS->{mta}->SendMessageString($this, $bouncetext, '<>')
+      or MailScanner::Log::WarnLog(
+           "Could not send sender MCP bounce attachment, %s", $!);
+  } else {
+    MailScanner::Log::WarnLog(
+      "Failed to create sender MCP bounce attachment, %s", $!);
+  }
+}
+
 
 # We want to send a message to the recipient saying that their spam
 # mail has not been delivered.
@@ -484,6 +569,7 @@ sub HandleMCPNotify {
 
   my($from,$to,$subject,$date,$spamreport,$hostname,$day,$month,$year);
   my($emailmsg, $line, $messagefh, $filename, $localpostmaster, $id);
+  my($postmastername);
 
   $from = $this->{from};
 
@@ -513,15 +599,17 @@ sub HandleMCPNotify {
   # Setup other variables they can use in the message template
   $id = $this->{id};
   $localpostmaster = MailScanner::Config::Value('localpostmaster', $this);
+  $postmastername  = MailScanner::Config::LanguageValue($this, 'mailscanner');
   $hostname = MailScanner::Config::Value('hostname', $this);
   $subject = $this->{subject};
-  $date = scalar localtime;
-  $spamreport = $this->{spamreport};
+  $date = $this->{datestring}; # scalar localtime;
+  $spamreport = $this->{mcpreport};
   # And let them put the date number in there too
-  ($day, $month, $year) = (localtime)[3,4,5];
-  $month++;
-  $year += 1900;
-  my $datenumber = sprintf("%04d%02d%02d", $year, $month, $day);
+  #($day, $month, $year) = (localtime)[3,4,5];
+  #$month++;
+  #$year += 1900;
+  #my $datenumber = sprintf("%04d%02d%02d", $year, $month, $day);
+  my $datenumber = $this->{datenumber};
 
 
   my($to, %tolist);
@@ -550,7 +638,7 @@ sub HandleMCPNotify {
     # Boring untainting again...
     /(.*)/;
     $line = eval "\"$1\"";
-    $emailmsg .= $line . "\n";
+    $emailmsg .= MailScanner::Config::DoPercentVars($line) . "\n";
   }
   $messagefh->close();
 
