@@ -2,7 +2,7 @@
 #   MailScanner - SMTP E-Mail Virus Scanner
 #   Copyright (C) 2002  Julian Field
 #
-#   $Id: Sendmail.pm,v 1.56.2.22 2004/06/01 16:09:59 jkf Exp $
+#   $Id: Sendmail.pm,v 1.56.2.33 2005/03/27 18:21:32 jkf Exp $
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -40,13 +40,15 @@ use DirHandle;
 use vars qw($VERSION);
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = substr q$Revision: 1.56.2.22 $, 10;
+$VERSION = substr q$Revision: 1.56.2.33 $, 10;
 
 # Command-line options you need to give to sendmail to sensibly process a
 # message that is piped to it. Still need to add "-f" for specifying the
 # envelope sender address. This is usually local postmaster.
 my $SendmailOptions = "-t -oi -oem -F MailScanner -f";
 my $UnsortedBatchesLeft;
+# This is true if the queue dir contains qf,df,xf sub-dirs.
+my %IsNestedQueue;
 
 # Attributes are
 #
@@ -66,6 +68,9 @@ sub initialise {
   MailScanner::Config::Default('sendmail2',
                                MailScanner::Config::Value('sendmail'));
   $UnsortedBatchesLeft = 0; # Disable queue-clearing mode
+  #%MailScanner::Sendmail::IsNestedQueue = ();
+  #print STDERR "Clearing IsNestedQueue\n";
+  %IsNestedQueue = ();
 }
 
 # Constructor.
@@ -232,19 +237,25 @@ my($sed) = "/bin/sed";
 #MailScanner::Log::InfoLog("Configuring mailscanner for sendmail...");
 
   sub DFileName {
-    my($this, $id) = @_;
+    my($this, $id, $dir) = @_;
+    #print STDERR "df IsNested $dir = " . $IsNestedQueue{$dir} . "\n";
+    return "df/df$id" if $IsNestedQueue{$dir};
     return "df$id";
   }
 
   # No change for V4
   sub HFileName {
-    my($this, $id) = @_;
+    my($this, $id, $dir) = @_;
+    #print STDERR "qf IsNested $dir = " . $IsNestedQueue{$dir} . "\n";
+    return "qf/qf$id" if $IsNestedQueue{$dir};
     return "qf$id";
   }
 
   # No change for V4
   sub TFileName {
-    my($this, $id) = @_;
+    my($this, $id, $dir) = @_;
+    #print STDERR "tf IsNested $dir = " . $IsNestedQueue{$dir} . "\n";
+    return "tf/tf$id" if $IsNestedQueue{$dir};
     return "tf$id";
   }
 
@@ -259,7 +270,7 @@ my($sed) = "/bin/sed";
 
     my($RQf) = $message->{store}{inhhandle};
 
-    my($InHeader, @results, $msginfo, $from);
+    my($InHeader, $InSubject, @results, $msginfo, $from);
     my($ip, $Rline);
     my($Line, $Flags);
     my($RFound, $SFound, $IPFound);
@@ -274,6 +285,7 @@ my($sed) = "/bin/sed";
     # between me opening it and locking it.
     seek($RQf, 0, 0);
     $InHeader = 0;
+    $InSubject = 0;
     while(<$RQf>) {
       last if /^\./; # Bat book section 23.9.19
       chomp; # Chomp everything now. We can easily add it back later.
@@ -313,12 +325,16 @@ my($sed) = "/bin/sed";
         }
         $IPFound = 1; # We have found the client IP address
       }
-      $InHeader = 1 if $Line =~ /^H/;
+      $InSubject = 0, $InHeader = 1 if $Line =~ /^H/;
       if ($Line !~ /^[H\t ]/) {
         $InHeader = 0;
+        $InSubject = 0;
         push @{$message->{metadata}}, $_; # Put non-headers into @metadata
         next;
       }
+      if ($InSubject && $Line =~ /^\s/) {
+        $message->{subject} .= $Line;
+      }
       $Line =~ s/^H//;
       # JKF 18/04/2001 Delete ?flags? for 0 or more flags for sendmail 8.11
       $Line =~ s/^(\?[^?]*\?)//;
@@ -329,14 +345,18 @@ my($sed) = "/bin/sed";
         # JKF $Line =~ s/[\x80-\xff]/\$/g;
       }
       push @{$message->{headers}}, $Line;
-      if ($Line =~ /^Subject:\s+(\S.*)$/i) {
+      if ($Line =~ /^Subject:\s*(\S.*)?$/i) {
         $message->{subject} = $1;
+        $InSubject = 1;
       }
     }
     # Not every qf file defined an IP address if it is a bounce.
     # So provide an IP if we haven't found one.
     $message->{clientip} = '0.0.0.0' unless $IPFound;
 
+    # Decode the ISO encoded Subject line
+    $message->{subject} = MIME::WordDecoder::unmime($message->{subject});
+
     # Every qf file should at least define the sender, 1 recipient and the
     # IP address. Everything else is optional, and is preserved as
     # MailScanner may not understand all the types of line.
@@ -399,6 +419,27 @@ my($sed) = "/bin/sed";
     }
   }
 
+  sub UniqHeader {
+    my($this, $message, $key) = @_;
+
+    my $linenum;
+    my $foundat = -1;
+    for ($linenum=0; $linenum<@{$message->{metadata}}; $linenum++) {
+      next unless $message->{metadata}[$linenum] =~ /^H(\?[^?]*\?)?$key/i;
+
+      # Have found the header line, skip it if we haven't seen it before
+      ($foundat = $linenum), next if $foundat == -1;
+
+      # Have found the right line
+      splice(@{$message->{metadata}}, $linenum, 1);
+      # Now delete the continuation lines
+      while($message->{metadata}[$linenum] =~ /^\s/) {
+        splice(@{$message->{metadata}}, $linenum, 1);
+      }
+      $linenum--; # Allow for 2 neighbouring instances of $key
+    }
+  }
+
   sub ReplaceHeader {
     my($this, $message, $key, $newvalue) = @_;
 
@@ -461,7 +502,7 @@ my($sed) = "/bin/sed";
     }
 
     $message->{metadata}[$oldlocation] =~
-      s/^H(\?[^?]*\?)?$key\s+/H$1$key $newvalue$sep/i;
+      s/^H(\?[^?]*\?)?$key\s*/H$1$key $newvalue$sep/i;
   }
 
   sub TextStartsHeader {
@@ -635,6 +676,22 @@ my($sed) = "/bin/sed";
   }
 
 
+  # Append, add or replace a given header with a given value.
+  sub AddMultipleHeaderName {
+    my $this = shift;
+    my($message, $headername, $headervalue, $separator) = @_;
+
+    my($multiple) = MailScanner::Config::Value('multipleheaders', $message);
+    $this->AppendHeader ($message, $headername, $headervalue, $separator)
+      if $multiple eq 'append';
+
+    $this->AddHeader    ($message, $headername, $headervalue)
+      if $multiple eq 'add';
+
+    $this->ReplaceHeader($message, $headername, $headervalue)
+      if $multiple eq 'replace';
+  }
+
   # Append, add or replace a given header with a given value.
   sub AddMultipleHeader {
     my $this = shift;
@@ -763,37 +820,79 @@ my($sed) = "/bin/sed";
       $invalidfiles = "";
 
       # Loop through each of the inq directories
-      foreach $queuedirname (@queuedirnames) {
+      # Patch to combat starving in emergency queue mode
+      # foreach $queuedirname (@queuedirnames) {
+      my @aux_queuedirnames=@queuedirnames;
+      while( defined($queuedirname=splice(@aux_queuedirnames,
+            ($UnsortedBatchesLeft<=0 ? 0 :int(rand(@aux_queuedirnames))),1))) {
         #print STDERR "Scanning dir $queuedirname\n";
-        unless (chdir $queuedirname) {
-          MailScanner::Log::WarnLog("Cannot cd to dir %s to read messages, %s",
-                       $queuedirname, $!);
-          next;
-        }
-
-        $queuedir->open('.')
-          or MailScanner::Log::DieLog("Cannot open queue dir %s for reading " .
-                                      "message batch, %s", $queuedirname, $!);
-        $mta = $global::MS->{mta};
-        #print STDERR "Searching " . $queuedirname . " for messages\n";
-
-        # Read in modification dates of the qf files & use them in date order
-        while(defined($file = $queuedir->read())) {
-          # Optimised by binning the 50% that aren't H files first
-          next unless $file =~ /$mta->{HFileRegexp}/;
-          #print STDERR "Found message file $file\n";
-          $MsgsInQueue++; # Count the size of the queue
-          push @SortedFiles, "$queuedirname/$file";
-          if ($UnsortedBatchesLeft<=0) {
-            # Running normally
-            $tmpdate = (stat($file))[9]; # 9 = mtime
-            next unless -f _;
-            next if -z _; # Skip 0-length qf files
-            $ModDate{"$queuedirname/$file"} = $tmpdate; # Push msg into list
-            #print STDERR "Stored message file $file\n";
+        #print STDERR "IsNestedQueue = " . join(',',%IsNestedQueue) . "\n";
+        if ($IsNestedQueue{$queuedirname}) {
+          # Queue is nested, so $queuedirname ends with /qf
+          #print STDERR "$queuedirname is nested\n";
+          $queuedirname .= '/qf';
+          unless (chdir $queuedirname) {
+            MailScanner::Log::WarnLog("Cannot cd to dir %s to read messages, %s",
+                         $queuedirname, $!);
+            next;
           }
+          $queuedir->open('.')
+            or MailScanner::Log::DieLog("Cannot open queue dir %s for " .
+                                        "reading message batch, %s",
+                                        $queuedirname, $!);
+          $mta = $global::MS->{mta};
+          #print STDERR "Searching " . $queuedirname . " for messages\n";
+
+          # Read in modification dates of the qf files & use them in date order
+          while(defined($file = $queuedir->read())) {
+            #print STDERR "Found $file\n";
+            # Optimised by binning the 50% that aren't H files first
+            next unless $file =~ /$mta->{HFileRegexp}/;
+            #print STDERR "Found message file $file\n";
+            $MsgsInQueue++; # Count the size of the queue
+            push @SortedFiles, "$queuedirname/$file";
+            if ($UnsortedBatchesLeft<=0) {
+              # Running normally
+              $tmpdate = (stat($file))[9]; # 9 = mtime
+              next unless -f _;
+              next if -z _; # Skip 0-length qf files
+              $ModDate{"$queuedirname/$file"} = $tmpdate; # Push msg into list
+              #print STDERR "Stored message file $file\n";
+            }
+          }
+          $queuedir->close();
+        } else {
+          unless (chdir $queuedirname) {
+            MailScanner::Log::WarnLog("Cannot cd to dir %s to read messages, %s",
+                         $queuedirname, $!);
+            next;
+          }
+          $queuedir->open('.')
+            or MailScanner::Log::DieLog("Cannot open queue dir %s for " .
+                                        "reading message batch, %s",
+                                        $queuedirname, $!);
+          $mta = $global::MS->{mta};
+          #print STDERR "Searching " . $queuedirname . " for messages\n";
+
+          # Read in modification dates of the qf files & use them in date order
+          while(defined($file = $queuedir->read())) {
+            # Optimised by binning the 50% that aren't H files first
+            next unless $file =~ /$mta->{HFileRegexp}/;
+            #print STDERR "Found message file $file\n";
+            $MsgsInQueue++; # Count the size of the queue
+            push @SortedFiles, "$queuedirname/$file";
+            if ($UnsortedBatchesLeft<=0) {
+              # Running normally
+              $tmpdate = (stat($file))[9]; # 9 = mtime
+              next unless -f _;
+              next if -z _; # Skip 0-length qf files
+              $ModDate{"$queuedirname/$file"} = $tmpdate; # Push msg into list
+              #print STDERR "Stored message file $file\n";
+            }
+          }
+          $queuedir->close();
         }
-        $queuedir->close();
+
       }
 
       # Not sorting the queue will save us considerably more time than
@@ -829,11 +928,13 @@ my($sed) = "/bin/sed";
         # same expression as $file [mumble mumble grrr mumble mumble]
         #print STDERR "Reading file $file from list\n";
         # Split pathname into dir and file again
+        # This now handles optional qf/ in the file for nested queues
         ($queuedirname, $file) = ($1,$2) if $file =~ /^(.*)\/([^\/]+)$/;
         next unless $file =~ /$mta->{HFileRegexp}/;
         $id = $1;
+        $queuedirname =~ s/\/qf$//;
 
-        #print STDERR "Adding $id to batch\n";
+        #print STDERR "Adding $id to batch from $queuedirname\n";
 
         # Lock and read the qf file. Skip this message if the lock fails.
         $newmessage = MailScanner::Message->new($id, $queuedirname);
@@ -846,7 +947,8 @@ my($sed) = "/bin/sed";
         $batch->{messages}{"$id"} = $newmessage;
         $batchempty = 0;
 
-        if (MailScanner::Config::Value("virusscan", $newmessage)) {
+        if (MailScanner::Config::Value("virusscan", $newmessage) =~ /1/ ||
+            MailScanner::Config::Value("dangerscan", $newmessage) =~ /1/) {
           $newmessage->NeedsScanning(1);
           $DirtyMsgs++;
           $DirtyBytes += $newmessage->{size};
@@ -933,26 +1035,36 @@ sub OriginalMsgHeaders {
 # sign that sendmail is running queue groups, which MailScanner cannot
 # handle.
 #
+# Update 8/12/2004: now support queue dirs that contain qf,df,xf subdirs
+# so this script has to return true in this case, but remember the fact
+# that the directory was nested.
+#
 # Called from main mailscanner script
 #
 sub CheckQueueIsFlat {
   my($dir) = @_;
 
   #MailScanner::Log::WarnLog("In CheckQueueIsFlat, dir is %s", $dir);
-  my($dirhandle, $f);
+  my($dirhandle, $f, $FoundQf, $FoundDf);
 
   $dirhandle = new DirHandle;
   $dirhandle->open($dir)
     or MailScanner::Log::DieLog("Cannot read queue directory $dir");
 
   # Check there are no q\d or qf subdirectories
+  $FoundQf = 0;
+  $FoundDf = 0;
   while($f = $dirhandle->read()) {
     # 2nd half of the line for SuSE Linux setups which put .hoststat
     # directory inside the queue!
     next if $f =~ /^\.\.?$/ || $f =~ /^\.hoststat/;
     # Delete core files
     $f =~ /^core$/ and unlink "core";
+    $FoundQf = 1 if $f eq 'qf' && -d "$dir/$f";
+    $FoundDf = 1 if $f eq 'df' && -d "$dir/$f";
+    next if $f =~ /^[qdxt]f$/; # These are allowed
     next unless $f =~ /^q[0-9f]$/;
+    # Now must allow for qf, df, etc directories.
     # Also needs untaint... sledgehammer. nut.
     $f =~ /(.*)/;
     MailScanner::Log::DieLog("Queue directory %s cannot contain sub-" .
@@ -962,6 +1074,12 @@ sub CheckQueueIsFlat {
   }
   $dirhandle->close();
 
+  # Remember the dir was nested if necessary
+  $IsNestedQueue{$dir} = ($FoundQf && $FoundDf)?1:0;
+  #print STDERR "Set IsNestedQueue for $dir so now " . join(',',%IsNestedQueue) . "\n";
+  MailScanner::Log::InfoLog("Queue directory %s is nested", $dir)
+    if $FoundQf && $FoundDf;
+
   return 1;
 }
 
